import logging
from dataclasses import dataclass
from enum import StrEnum
from functools import cached_property
from typing import Iterable

import funcy
from rdflib import URIRef
from rich.columns import Columns
from textual.containers import Vertical, VerticalScroll
from textual.widget import Widget
from textual.widgets import Static

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES


class TermStatus(StrEnum):
    """Status of an ontology term."""

    STABLE = 'stable'
    ARCHAIC = 'archaic'
    TESTING = 'testing'
    UNSTABLE = 'unstable'


@dataclass
class TermAndStatus:
    term: URIRef
    status: TermStatus


class TermsContent(VerticalScroll):
    """Display grouped list of terms."""

    DEFAULT_CSS = """
    TermsContent {
        padding: 1;
    }
    """


class OntologyFacet(Facet[Widget]):
    """Render an ontology."""

    @cached_property
    def grouped_terms(self) -> dict[NotLiteralNode | None, list[TermAndStatus]]:
        """Group terms by VANN categories."""
        rows = self.stored_query('terms.sparql', iri=self.this)
        grouped = [
            (
                row.get('group'),
                TermAndStatus(
                    term=row['term'],
                    status=status,
                ),
            )
            for row in rows
            if (
                status := TermStatus(
                    (status_literal := row.get('status'))
                    and status_literal.value
                    or 'stable',
                )
            ) != TermStatus.ARCHAIC
        ]

        return funcy.group_values(grouped)

    def show(self) -> Widget:
        """Render widget."""

        index_title = self.render(
            URIRef('https://iolanta.tech/visualizations/index.yaml'),
            as_datatype=DATATYPES.title,
        )
        self.logger.info('Index Retrieved: %s', index_title)

        vocabs = funcy.lpluck(
            'vocab',
            self.stored_query('visualization-vocab.sparql', iri=self.this),
        )

        for vocab in vocabs:
            vocab_title = self.render(
                vocab,
                as_datatype=DATATYPES.title,
            )
            self.logger.info(
                'Visualization vocabulary retrieved: %s',
                vocab_title,
            )

        return Vertical(
            PageTitle(self.this),
            TermsContent(
                Static(
                    Columns(
                        self._stream_columns(),
                        padding=(1, 2),
                    ),
                ),
            ),
        )

    def _stream_columns(self) -> Iterable[str]:
        for group, rows in self.grouped_terms.items():
            rendered_terms = '\n'.join([
                self.render(
                    row.term,
                    as_datatype=URIRef('https://iolanta.tech/cli/link'),
                )
                for row in rows
            ])

            if group is None:
                yield rendered_terms

            else:
                group_title = self.render(
                    group,
                    as_datatype=DATATYPES.title,
                )

                yield f'[b]{group_title}[/b]\n\n{rendered_terms}'



from pathlib import Path
from typing import Annotated

from fastmcp import FastMCP
from rdflib import URIRef
from yarl import URL

from iolanta.cli.main import render_and_return

mcp = FastMCP("Iolanta MCP Server")


@mcp.tool()
def render_uri(
    uri: Annotated[str, "URL, or file system path, to render"],
    as_format: Annotated[
        str, "Format to render as. Examples: `labeled-triple-set`, `mermaid`"
    ],
) -> str:
    """Render a URI."""
    # Parse string URL to URIRef
    node_url = URL(uri)
    if node_url.scheme and node_url.scheme != "file":
        node = URIRef(uri)
    else:
        path = Path(node_url.path).absolute()
        node = URIRef(f"file://{path}")

    return str(render_and_return(node=node, as_datatype=as_format))


def app():
    mcp.run()


if __name__ == "__main__":
    app()



from pathlib import Path
from typing import cast

from more_itertools import consume
from rdflib import Graph, Literal
from rich.table import Table

from iolanta.cli.formatters.pretty import pretty_print_value
from iolanta.facets.cli.base import Renderable, RichFacet
from iolanta.facets.errors import NotALiteral

META = Path(__file__).parent / 'data' / 'query_result.yamlld'


class ConstructResultTableFacet(RichFacet):
    """Render CONSTRUCT query results as table."""

    META = META

    def show(self) -> Renderable:
        """Render Graph (CONSTRUCT result) as table."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(node=self.this)

        graph = cast(Graph, self.this.value)

        if not graph:
            table = Table(
                'Subject',
                'Predicate',
                'Object',
                show_header=True,
                header_style="bold magenta",
            )
            return table

        table = Table(
            'Subject',
            'Predicate',
            'Object',
            show_header=True,
            header_style="bold magenta",
        )

        consume(
            table.add_row(
                *[
                    str(pretty_print_value(value))
                    for value in triple
                ],
            )
            for triple in graph
        )

        return table



import rdflib

LOCAL = rdflib.Namespace('local:')
IOLANTA = rdflib.Namespace('https://iolanta.tech/')
DATATYPES = rdflib.Namespace('https://iolanta.tech/datatypes/')
IOLANTA_FACETS = rdflib.Namespace('pkg:pypi/iolanta#')
NP = rdflib.Namespace('https://www.nanopub.org/nschema#')
RDFG = rdflib.Namespace('https://www.w3.org/2004/03/trix/rdfg-1/')
SDO = rdflib.SDO

META = rdflib.URIRef('iolanta://_meta')


class DC(rdflib.DC):
    _NS = rdflib.Namespace('https://purl.org/dc/elements/1.1/')


class OWL(rdflib.OWL):
    _NS = rdflib.Namespace('https://www.w3.org/2002/07/owl#')


class RDFS(rdflib.RDFS):
    _NS = rdflib.Namespace('http://www.w3.org/2000/01/rdf-schema#')


class RDF(rdflib.RDF):
    _NS = rdflib.Namespace('http://www.w3.org/1999/02/22-rdf-syntax-ns#')


class DCTERMS(rdflib.DCTERMS):
    _NS = rdflib.Namespace('https://purl.org/dc/terms/')


class VANN(rdflib.VANN):
    ...


class FOAF(rdflib.FOAF):
    ...


class XSD(rdflib.XSD):
    _NS = rdflib.Namespace('https://www.w3.org/2001/XMLSchema#')


class PROV(rdflib.PROV):
    _NS = rdflib.Namespace('https://www.w3.org/ns/prov#')



from dataclasses import dataclass

from rdflib import URIRef

from iolanta.models import NotLiteralNode


@dataclass
class Location:
    """Unique ID and IRI associated with it."""

    page_id: str
    url: NotLiteralNode
    facet_iri: URIRef | None = None



import locale
import logging
import sys
from pathlib import Path
from typing import Annotated

import loguru
import platformdirs
from documented import DocumentedError
from rdflib import Graph, Literal, URIRef
from rich.console import Console
from rich.markdown import Markdown
from rich.table import Table
from typer import Argument, Exit, Option, Typer
from yarl import URL

from iolanta.cli.models import LogLevel
from iolanta.facets.errors import FacetNotFound
from iolanta.iolanta import Iolanta
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES
from iolanta.query_result import (
    QueryResult,
    SPARQLParseException,
    SelectResult,
)

DEFAULT_LANGUAGE = locale.getlocale()[0].split("_")[0]


console = Console()


app = Typer(no_args_is_help=True)


def string_to_node(name: str) -> NotLiteralNode:
    """
    Parse a string into a node identifier.

    String might be:
      * a URL,
      * or a local disk path.
    """
    url = URL(name)
    if url.scheme:
        return URIRef(name)

    path = Path(name).absolute()
    return URIRef(f"file://{path}")


def decode_datatype(datatype: str) -> URIRef:
    if datatype.startswith("http"):
        return URIRef(datatype)

    return URIRef(f"https://iolanta.tech/datatypes/{datatype}")


def setup_logging(log_level: LogLevel):
    """Configure and return logger."""
    level = {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
    }[log_level]

    log_file_path = (
        platformdirs.user_log_path(
            "iolanta",
            ensure_exists=True,
        )
        / "iolanta.log"
    )

    level_name = {
        logging.DEBUG: "DEBUG",
        logging.INFO: "INFO",
        logging.WARNING: "WARNING",
        logging.ERROR: "ERROR",
    }[level]

    loguru.logger.remove()
    loguru.logger.add(
        log_file_path,
        level=level_name,
        format="{time} {level} {message}",
        enqueue=True,
    )
    loguru.logger.add(
        sys.stderr,
        level=level_name,
        format="{time} | {level:<8} | {name}:{function}:{line} - {message}",
    )
    loguru.logger.level(level_name)

    return loguru.logger


def handle_error(
    error: Exception,
    log_level: LogLevel,
    use_markdown: bool = True,
) -> None:
    """
    Handle an error by checking log level and printing appropriately.

    If log level is DEBUG or INFO, re-raise the error.
    Otherwise, print it and exit with code 1.
    """
    level = {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
    }[log_level]

    if level in {logging.DEBUG, logging.INFO}:
        raise error

    if use_markdown:
        console.print(
            Markdown(
                str(error),
                justify="left",
            ),
        )
    else:
        console.print(str(error))

    raise Exit(1)


def create_query_node(query_result: QueryResult) -> Literal:
    """
    Create a Literal node from a query result.

    Converts QueryResult (SelectResult, Graph, or bool) into a Literal
    with the appropriate datatype for facet rendering.
    """
    match query_result:
        case SelectResult():
            return Literal(
                query_result,
                datatype=DATATYPES["sparql-select-result"],
            )
        case Graph():
            return Literal(
                query_result,
                datatype=DATATYPES["sparql-construct-result"],
            )
        case bool():
            return Literal(
                query_result,
                datatype=DATATYPES["sparql-ask-result"],
            )


def render_and_return(
    node: Literal | URIRef,
    as_datatype: str,
    language: str = DEFAULT_LANGUAGE,
    log_level: LogLevel = LogLevel.ERROR,
):
    """
    Render a node.

    The node must be either a URIRef (for URLs) or a Literal (for query results).
    """
    logger = setup_logging(log_level)

    # Determine Iolanta instance based on node type
    if isinstance(node, Literal):
        # Literal nodes (e.g., from query results) are used directly
        # Use current directory as project_root for query results
        iolanta: Iolanta = Iolanta(
            language=Literal(language),
            logger=logger,
            project_root=Path.cwd(),
        )
    elif isinstance(node, URIRef):
        # URIRef - determine project_root if it's a file:// URI
        if str(node).startswith("file://"):
            path = Path(str(node).replace("file://", ""))
            iolanta: Iolanta = Iolanta(
                language=Literal(language),
                logger=logger,
                project_root=path,
            )
        else:
            iolanta: Iolanta = Iolanta(
                language=Literal(language),
                logger=logger,
            )
    else:
        # This should never happen due to type checking, but kept for safety
        raise TypeError(f"Expected Literal or URIRef, got {type(node)}")

    return iolanta.render(
        node=node,
        as_datatype=decode_datatype(as_datatype),
    )


@app.command(name="render")
def render_command(  # noqa: WPS231, WPS238, WPS210, C901
    url: Annotated[str | None, Argument()] = None,
    query: Annotated[
        str | None,
        Option(
            "--query",
            help="SPARQL query to execute.",
        ),
    ] = None,
    as_datatype: Annotated[
        str | None,
        Option(
            "--as",
        ),
    ] = None,
    language: Annotated[
        str,
        Option(
            help="Data language to prefer.",
        ),
    ] = DEFAULT_LANGUAGE,
    log_level: LogLevel = LogLevel.ERROR,
):
    """Render a given URL."""
    if query is not None:
        # For queries, default to 'table' format
        if as_datatype is None:
            as_datatype = "table"

        # Setup logging and create Iolanta instance (unlikely to raise)
        logger = setup_logging(log_level)
        iolanta: Iolanta = Iolanta(
            language=Literal(language),
            logger=logger,
            project_root=Path.cwd(),
        )

        try:
            renderable = render_and_return(
                node=create_query_node(iolanta.query(query)),
                as_datatype=as_datatype,
                language=language,
                log_level=log_level,
            )
        except (SPARQLParseException, DocumentedError, FacetNotFound) as error:
            handle_error(error, log_level, use_markdown=True)
        except Exception as error:
            handle_error(error, log_level, use_markdown=False)
        else:
            # FIXME: An intermediary Literal can be used to dispatch rendering.
            match renderable:
                case Table() as table:
                    console.print(table)

                case unknown:
                    console.print(unknown)
        return

    if url is None:
        console.print("Error: Either URL or --query must be provided")
        raise Exit(1)

    # For URLs, default to interactive mode
    if as_datatype is None:
        as_datatype = "https://iolanta.tech/cli/interactive"

    # Parse string URL to URIRef (URL() is permissive and won't raise)
    node_url = URL(url)
    if node_url.scheme and node_url.scheme != "file":
        node = URIRef(url)
    else:
        path = Path(node_url.path).absolute()
        node = URIRef(f"file://{path}")

    try:
        renderable = render_and_return(
            node=node,
            as_datatype=as_datatype,
            language=language,
            log_level=log_level,
        )
    except DocumentedError as error:
        handle_error(error, log_level, use_markdown=True)
    except Exception as error:
        handle_error(error, log_level, use_markdown=False)
    else:
        # FIXME: An intermediary Literal can be used to dispatch rendering.
        match renderable:
            case Table() as table:
                console.print(table)

            case unknown:
                console.print(unknown)



from pathlib import Path
from typing import Annotated, Iterable
from typing import Literal as TypingLiteral

from pydantic import (
    AnyUrl,
    BaseModel,
    Field,
    TypeAdapter,
    field_serializer,
    field_validator,
    validator,
)
from rdflib import BNode, Literal, Node, URIRef

from iolanta import Facet
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES


class WithFeedback(BaseModel):
    feedback: Annotated[list[str], Field(default_factory=list)]


class LabeledURI(WithFeedback, BaseModel):
    type: TypingLiteral['uri'] = 'uri'
    uri: AnyUrl
    label: str

    def construct_feedback(self) -> Iterable[str]:
        if str(self.label) == str(self.uri):
            yield (
                'For this URI, the label is the same as the URI. We were '
                'unable to render that URI.'
            )


class LabeledBlank(WithFeedback, BaseModel):
    type: TypingLiteral['blank'] = 'blank'
    identifier: str
    label: str


class LabeledLiteral(WithFeedback, BaseModel):
    type: TypingLiteral['literal'] = 'literal'
    value: str
    language: str | None
    datatype: str | None
    label: str


class LabeledTriple(BaseModel):
    subject: LabeledURI | LabeledBlank
    predicate: LabeledURI
    object_: LabeledURI | LabeledBlank | LabeledLiteral


def construct_uri_feedback(uri: AnyUrl, label: str) -> Iterable[str]:
    if str(uri) == str(label):
        yield (
            'For this URI, the label is the same as the URI. We were '
            'unable to render that URI.'
        )


def construct_blank_feedback(bnode, label) -> Iterable[str]:
    if str(bnode) == str(label):
        yield (
            'For this blank node, the label is the same as the blank node. '
            'We were unable to render that blank node.'
        )


def construct_literal_feedback(literal, label):
    if label.startswith('http'):
        yield (
            'This RDF literal seems to be actually a URL. Good chance is '
            'that it should not be a literal.'
        )
    
    elif ':' in label[:5]:
        yield (
            'This RDF literal seems to be actually a QName (prefixed URI). '
            'Good chance is that it should not be a literal.'
        )


class LabeledTripleSet(Facet[list[LabeledTriple]]):
    """A set of labeled triples."""

    META = Path(__file__).parent / 'data' / 'labeled_triple_set.yamlld'

    def render_label(self, node: NotLiteralNode) -> str:
        return self.render(node, as_datatype=DATATYPES.title)

    def parse_term(self, term: Node):
        match term:
            case URIRef() as uriref:
                uri = AnyUrl(uriref)
                label = self.render_label(uriref)

                return LabeledURI(
                    uri=uri,
                    label=label,
                    feedback=list(construct_uri_feedback(uri=uri, label=label)),
                )

            case BNode() as bnode:
                label = self.render_label(bnode)
                return LabeledBlank(
                    identifier=bnode,
                    label=label,
                    feedback=list(construct_blank_feedback(bnode=bnode, label=label)),
                )

            case Literal() as literal:
                label = self.render_label(literal)
                return LabeledLiteral(
                    value=literal,
                    language=literal.language,
                    datatype=literal.datatype,
                    label=label,
                    feedback=list(construct_literal_feedback(literal=literal, label=label)),
                )

    def show(self):
        rows = self.stored_query('triples.sparql', graph=self.this)
        triples = [
            LabeledTriple(
                subject=self.parse_term(row['subject']),
                predicate=self.parse_term(row['predicate']),
                object_=self.parse_term(row['object']),
            )
            for row in rows
        ]

        return TypeAdapter(list[LabeledTriple]).dump_json(triples, indent=2).decode()



from pathlib import Path

from more_itertools import consume, first
from rdflib import Literal
from rich.table import Table

from iolanta.cli.formatters.pretty import pretty_print_value
from iolanta.facets.cli.base import Renderable, RichFacet
from iolanta.facets.errors import NotALiteral

META = Path(__file__).parent / 'data' / 'query_result.yamlld'


class SelectResultTableFacet(RichFacet):
    """Render SELECT query results as table."""

    META = META

    def show(self) -> Renderable:
        """Render SelectResult as table."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(node=self.this)

        query_result = self.this.value

        if not query_result:
            table = Table(show_header=True, header_style="bold magenta")
            return table

        columns = first(query_result).keys()

        table = Table(
            *columns,
            show_header=True,
            header_style="bold magenta",
        )

        consume(
            table.add_row(
                *[
                    str(pretty_print_value(value))
                    for value in row.values()
                ],
            )
            for row in query_result
        )

        return table



from typing import Type

from rdflib import URIRef

from iolanta.facets.facet import Facet
from iolanta.resolvers.base import Resolver


class PythonImportResolver(Resolver):
    """Resolve facet IRIs into classes by importing Python modules."""

    def resolve(self, uri: URIRef) -> Type[Facet]:
        """Find a resolver by IRI in python:<module>.<class> format."""
        raise NotImplementedError(
            f'{uri} URL for facet import is unsupported anymore '
            f'due to security reasons.',
        )



import functools
import itertools
from pathlib import Path
from typing import ClassVar, Iterable

import funcy
from rich.console import RenderResult
from rich.text import Text
from textual.binding import Binding, BindingType
from textual.containers import Vertical  # removed unused VerticalScroll (F401)
from textual.reactive import Reactive
from textual.widget import Widget
from textual.widgets import Label, ListItem, ListView

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.facets.textual_default.triple_uri_ref import TripleURIRef
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES, RDF

INSTANCE_RENDER_RADIUS = 50


class InstanceItem(ListItem):
    """An item in class instances list."""

    renderable: Reactive[RenderResult | None](None, init=True, layout=True)

    def __init__(
        self,
        node: NotLiteralNode,
        parent_class: NotLiteralNode,
    ):
        """Specify the node, its class, and that we are not rendered yet."""
        self.node = node
        self.parent_class = parent_class
        self.renderable = None
        super().__init__()

    def render(self) -> RenderResult:
        """
        Render this class instance.

        Render either the result or the raw node.

        # FIXME Calculate QName at least. Or CURIE? Or whatsit?
        """
        qname = self.app.iolanta.node_as_qname(self.node)
        return self.renderable or Text(
            f'⏳ {qname}',
            style='#696969',
        )

    def update(self, renderable: RenderResult):
        """
        Assign the render result.

        A separate method is needed for this because we call it from a thread.
        """
        self.renderable = renderable
        self.refresh()


class InstancesList(ListView):   # noqa: WPS214
    """Instances list."""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('enter', 'goto', 'Goto'),
        Binding('p', 'provenance', 'Provenan©e'),
    ]

    DEFAULT_CSS = """
    InstancesList {
        height: 1fr;
        layout: vertical;
        overflow-y: auto;
    }
    """

    FIRST_CHUNK_SIZE = 15
    DEFAULT_CHUNK_SIZE = 10

    def __init__(
        self,
        instances: Iterable[NotLiteralNode],
        parent_class: NotLiteralNode,
    ):
        """Specify the instances to render and their class."""
        self.instances = instances
        self.parent_class = parent_class
        super().__init__()

    def render_newly_added_instances(self, instance_items: list[InstanceItem]):
        """
        Render each of the given instance items.

        Must be called in a worker.
        """
        for instance_item in instance_items:
            self.app.call_from_thread(
                instance_item.update,
                self.app.iolanta.render(
                    instance_item.node,
                    as_datatype=DATATYPES.title,
                ),
            )

    def stream_instance_items_chunk(
        self,
        count: int | None = None,
    ) -> Iterable[InstanceItem]:
        """Return a chunk of unique class instances."""
        chunk = itertools.islice(
            self.instances,
            count or self.DEFAULT_CHUNK_SIZE,
        )
        for instance in chunk:  # noqa: WPS526
            yield InstanceItem(
                node=instance,
                parent_class=self.parent_class,
            )

    def action_goto(self):
        """Navigate."""
        self.app.action_goto(self.highlighted_child.node)

    def action_provenance(self):
        """Navigate to provenance for the property value."""
        self.app.action_goto(
            TripleURIRef.from_triple(
                subject=self.highlighted_child.node,
                predicate=RDF.type,
                object_=self.parent_class,
            ),
        )

    def compose(self):
        """Fill in the instances."""
        for instance in self.instances:   # noqa: WPS526
            yield InstanceItem(
                node=instance,
                parent_class=self.parent_class,
            )

    def render_instances(self):
        """Render all instances."""
        for instance_item in self.children:
            self.app.call_from_thread(
                instance_item.update,
                self.app.iolanta.render(
                    instance_item.node,
                    as_datatype=DATATYPES.title,
                ),
            )

    def on_mount(self):
        """Render the first chunk of instances."""
        self.run_worker(
            self.render_instances,
            thread=True,
        )


class Bottom(Label):
    """Label below the instances list."""

    DEFAULT_CSS = """
    Bottom {
        padding-top: 1;
        padding-bottom: 1;
        dock: bottom;
    }
    """


class InstancesBody(Vertical):
    """Container for instances list and accompanying bells and whistles."""

    DEFAULT_CSS = """
    InstancesBody {
        height: auto;
        max-height: 100%;
    }
    """


class Class(Facet[Widget]):
    """Render instances of a class."""

    META = Path(__file__).parent / 'textual-class.yamlld'

    def stream_instances(self) -> Iterable[NotLiteralNode]:
        """
        Query and stream class instances lazily.

        The operation of rendering an instance is not pure: it might cause us
        to retrieve more data and load said data into the graph. That's because
        we do multiple query attempts.

        We have to stop if a subsequent attempt returns no results. That's why
        we can't use `funcy.distinct()` or something similar.
        """
        return set(
            funcy.pluck(
                'instance',
                self.stored_query('instances.sparql', this=self.this),
            ),
        )

    def show(self) -> Widget:
        """Render the instances list."""
        return InstancesBody(
            PageTitle(self.this),
            InstancesList(
                instances=list(self.stream_instances()),
                parent_class=self.this,
            ),
        )



import csv
import io
from pathlib import Path

from rdflib import Literal

from iolanta.facets.cli.base import Renderable, RichFacet
from iolanta.facets.errors import NotALiteral

META = Path(__file__).parent / 'data' / 'query_result.yamlld'


class SelectResultCsvFacet(RichFacet):
    """Render SELECT query results as CSV."""

    META = META

    def show(self) -> Renderable:
        """Render SelectResult as CSV."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(node=self.this)

        query_result = self.this.value

        if not query_result:
            return ""

        output = io.StringIO()
        first_row = query_result[0]
        fieldnames = first_row.keys()

        writer = csv.DictWriter(output, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(query_result)

        return output.getvalue()



from dataclasses import dataclass
from typing import Dict, List, Mapping, Optional, Union

from documented import DocumentedError
from frozendict import frozendict
from pyparsing import ParseException
from rdflib import Graph
from rdflib.term import Identifier, Node, Variable

SelectRow = Mapping[str, Node]


class SelectResult(List[SelectRow]):
    """Result of a SPARQL SELECT."""

    @property
    def first(self) -> Optional[SelectRow]:
        """Return first element of the list."""
        return self[0] if self else None


SPARQLQueryArgument = Optional[Union[Node, str, int, float]]


QueryResult = Union[
    SelectResult,   # SELECT
    Graph,          # CONSTRUCT
    bool,           # ASK
]


def format_query_bindings(
    bindings: List[Dict[Variable, Identifier]],
) -> SelectResult:
    """
    Format bindings before returning them.

    Converts Variable to str for ease of addressing.
    """
    return SelectResult([
        frozendict({
            str(variable_name): rdf_value
            for variable_name, rdf_value   # noqa: WPS361
            in row.items()
        })
        for row in bindings
    ])


@dataclass
class SPARQLParseException(DocumentedError):
    """
    SPARQL query is invalid.

    Error:

    ```
    {self.error}
    ```

    Query:
    ```sparql hl_lines="{self.highlight_code}"
    {self.query}
    ```
    """  # noqa: D412

    error: ParseException
    query: str

    @property
    def highlight_code(self):
        """Define lines to highlight."""
        return self.error.lineno



from pathlib import Path
from typing import Iterable

from rdflib import BNode, Literal, Node, URIRef

from iolanta import Facet
from iolanta.mermaid.models import (
    Diagram,
    MermaidBlankNode,
    MermaidEdge,
    MermaidLiteral,
    MermaidScalar,
    MermaidSubgraph,
    MermaidURINode,
)
from iolanta.namespaces import DATATYPES
from pydantic import AnyUrl
from rdflib import URIRef as RDFURIRef


class TaskNode(MermaidScalar):
    """A Mermaid node with attached CSS classes."""

    node: MermaidURINode | MermaidBlankNode | MermaidLiteral
    classes: list[str] = []

    def __str__(self) -> str:
        """Render the wrapped node."""
        return str(self.node)

    @property
    def id(self) -> str:
        """Get the ID of the wrapped node."""
        return self.node.id


class BlocksEdge(MermaidEdge):
    """
    {self.source.id} --> {self.target.id}
    """

    def __init__(self, source, target):
        # Initialize with empty title and a dummy predicate
        super().__init__(
            source=source,
            target=target,
            predicate=RDFURIRef('https://iolanta.tech/roadmap/blocks'),
            title='',
        )

    def __str__(self) -> str:
        # Override to remove intermediate node - just direct arrow
        return f'{self.source.id} --> {self.target.id}'


# Rebuild Pydantic models to resolve forward references
# Need to rebuild MermaidEdge first so MermaidSubgraph is available
MermaidEdge.model_rebuild()
BlocksEdge.model_rebuild()


class MermaidRoadmap(Facet[str]):
    """Mermaid roadmap diagram."""

    META = Path(__file__).parent / 'mermaid_roadmap.yamlld'

    inference_path = Path(__file__).parent / 'inference'

    def show(self) -> str:
        """Render mermaid roadmap diagram."""
        children = list(self.construct_mermaid_for_graph(self.this))
        
        # Extract class assignments from TaskNode instances
        tail_parts = ['classDef unblocked fill:#0a5,stroke:#063,stroke-width:2px,color:#fff;']
        for child in children:
            if isinstance(child, TaskNode) and child.classes:
                for class_name in child.classes:
                    tail_parts.append(f'class {child.id} {class_name}')
        
        tail = '\n'.join(tail_parts)
        
        return str(Diagram(
            children=children,
            tail=tail,
        ))

    def as_mermaid(self, node: Node):
        """Convert RDF node to Mermaid node."""
        match node:
            case URIRef() as uri:
                return MermaidURINode(
                    uri=uri,
                    url=AnyUrl(uri),
                    title=self.render(uri, as_datatype=DATATYPES.title),
                )
            case Literal() as literal:
                return MermaidLiteral(literal=literal)
            case BNode() as bnode:
                return MermaidBlankNode(
                    node=bnode,
                    title=self.render(bnode, as_datatype=DATATYPES.title),
                )
            case unknown:
                unknown_type = type(unknown)
                raise ValueError(f'Unknown node type: {unknown} ({unknown_type})')

    def construct_mermaid_for_graph(self, graph: URIRef) -> Iterable[MermaidScalar]:
        """Render graph as mermaid."""
        # Get nodes
        node_rows = self.stored_query('nodes.sparql')
        node_rows_list = list(node_rows)
        
        nodes = [
            TaskNode(
                node=self.as_mermaid(row['node']),
                classes=['unblocked'] if row.get('is_unblocked', False) else [],
            )
            for row in node_rows_list
        ]

        # Get edges for roadmap:blocks relationships
        edge_rows = self.stored_query('edges.sparql')
        edge_rows_list = list(edge_rows)
        
        edges = [
            BlocksEdge(
                source=self.as_mermaid(row['source']),
                target=self.as_mermaid(row['target']),
            )
            for row in edge_rows_list
        ]

        return [*nodes, *edges]



from pathlib import Path

from rdflib import Literal

from iolanta.facets.cli.base import Renderable, RichFacet
from iolanta.facets.errors import NotALiteral

META = Path(__file__).parent / 'data' / 'query_result.yamlld'


class AskResultTableFacet(RichFacet):
    """Render ASK query results as table/text."""

    META = META

    def show(self) -> Renderable:
        """Render bool (ASK result) as text."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(node=self.this)

        query_result = self.this.value

        return "✅ `True`" if query_result else "❌ `False`"



from collections import deque
from dataclasses import dataclass, field
from typing import Generic, TypeVar

LocationType = TypeVar('LocationType')


@dataclass
class NavigationHistory(Generic[LocationType]):
    """Navigation history."""

    past: deque = field(default_factory=deque)
    current: LocationType | None = None
    future: deque = field(default_factory=deque)

    def goto(self, location: LocationType) -> LocationType:
        """Go to a location."""
        if self.current is not None:
            self.past.append(self.current)
        self.current = location
        self.future.clear()

        return self.current

    def back(self) -> LocationType | None:
        """Go back in history."""
        if self.past:
            self.future.appendleft(self.current)
            self.current = self.past.pop()

        return self.current

    def forward(self) -> LocationType | None:
        """Go forward in history."""
        if self.future:
            self.past.append(self.current)
            self.current = self.future.popleft()

        return self.current



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from documented import DocumentedError

from iolanta.models import LDDocument


@dataclass
class NotAContext(DocumentedError):
    """
    This is not a context.

    Document is expected to be a context but it's not.

    It lacks the `@context` key.

    Document: {self.document}

    Path: {self.formatted_path}
    """

    document: LDDocument
    path: Optional[Path] = None

    @property
    def formatted_path(self):
        return self.path or '<Unknown>'


def ensure_is_context(document: LDDocument) -> LDDocument:
    if '@context' not in document:
        raise NotAContext(document=document)

    return document



from abc import abstractmethod
from typing import Type

from rdflib import URIRef

from iolanta.facets.facet import Facet


class Resolver:
    """Resolve facet IRIs into classes."""

    @abstractmethod
    def resolve(self, uri: URIRef) -> Type[Facet]:
        """Find a resolver by IRI."""



from typing import ClassVar

from rdflib import Literal, URIRef
from rdflib.term import Node
from rich.text import Text
from textual.app import ComposeResult, RenderResult
from textual.binding import Binding, BindingType
from textual.containers import Vertical, VerticalScroll
from textual.events import Click
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Static, TabbedContent, TabPane
from textual.worker import Worker, WorkerState

from iolanta.facets.textual_default.triple_uri_ref import TripleURIRef
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES


class PropertyName(Widget, can_focus=True, inherit_bindings=False):
    """Property name."""

    DEFAULT_CSS = """
    PropertyName {
        width: 15%;
        height: auto;
        margin-right: 1;
        text-style: bold;
    }

    PropertyName:hover {
        background: $boost;
    }

    PropertyName:focus {
        background: darkslateblue;
    }
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('enter', 'goto', 'Goto'),
    ]
    renderable: str | None = reactive[str | None](  # noqa: WPS465
        None,
        init=False,
        layout=True,
    )

    def __init__(
        self,
        iri: NotLiteralNode,
        qname: str,
    ):
        """Set the IRI."""
        self.iri = iri
        super().__init__()
        self.renderable = Text(  # noqa: WPS601
            f'⏳ {qname}',
            style='#696969',
        )

    def render_title(self):
        """Render title in a separate thread."""
        title = self.app.iolanta.render(self.iri, DATATYPES.title)
        return f'⤚{title}→'

    def render(self) -> RenderResult:
        """Render node title."""
        return self.renderable or '…'

    def on_worker_state_changed(self, event: Worker.StateChanged):
        """Show the title after it has been rendered."""
        match event.state:
            case WorkerState.SUCCESS:
                self.renderable = event.worker.result  # noqa: WPS601
                self.styles.color = 'white'

            case WorkerState.ERROR:
                raise ValueError(event)

    def action_goto(self):
        """Navigate."""
        self.app.action_goto(self.iri)

    def on_click(self, event: Click):
        """
        Navigate to the property if we are focused.

        TODO: Does not work; causes navigation even if not focused.
        """
        if self.has_focus:
            return self.action_goto()


class PropertyValue(Widget, can_focus=True, inherit_bindings=False):
    """
    Value of a property.

    Supports navigation and provenance.
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('enter', 'goto', 'Goto'),
        Binding('p', 'provenance', 'Provenan©e'),
    ]

    DEFAULT_CSS = """
    PropertyValue {
        width: auto;
        height: auto;
        padding-right: 1;
        padding-bottom: 1;
    }

    PropertyValue:hover {
        background: $boost;
    }

    PropertyValue:focus {
        background: darkslateblue;
    }
    """

    renderable: str | None = reactive[str | None](  # noqa: WPS465
        None,
        init=False,
        layout=True,
    )

    def __init__(
        self,
        property_value: Node,
        subject: NotLiteralNode,
        property_iri: NotLiteralNode,
        property_qname: str,
    ):
        """Initialize parameters for rendering, navigation, & provenance."""
        self.property_value = property_value
        self.subject = subject
        self.property_iri = property_iri
        super().__init__()
        self.renderable = Text(  # noqa: WPS601
            f'⏳ {property_value}',
            style='#696969',
        )

    @property
    def iri(self):
        """Return the property IRI for compatibility."""
        return self.property_value

    def render_title(self):
        """Render title in a separate thread."""
        return self.app.iolanta.render(
            self.iri,
            as_datatype=DATATYPES.title,
        )

    def on_worker_state_changed(self, event: Worker.StateChanged):
        """Show the title after it has been rendered."""
        match event.state:
            case WorkerState.SUCCESS:
                self.renderable = event.worker.result  # noqa: WPS601
                self.styles.color = 'white'

            case WorkerState.ERROR:
                raise ValueError(event)

    def render(self) -> RenderResult:
        """Render title of the node."""
        return self.renderable

    def action_goto(self):
        """Navigate."""
        self.app.action_goto(self.property_value)

    def action_provenance(self):
        """Navigate to provenance for the property value."""
        self.app.action_goto(
            TripleURIRef.from_triple(
                subject=self.subject,
                predicate=self.property_iri,
                object_=self.property_value,
            ),
        )

    def on_click(self, event: Click):
        """
        Navigate to the property if we are focused.

        FIXME: Does not work; causes navigation even if not focused.
        """
        if self.has_focus:
            return self.action_goto()


class PropertiesContainer(Vertical):
    """Contain all properties and their values."""

    DEFAULT_CSS = """
    PropertiesContainer {
        height: auto;
        padding: 1;
    }
    """

    def render_all_properties(self):
        """Render all property names & values."""
        widgets = self.query('PropertyName, PropertyValue')

        widget: PropertyName | PropertyValue
        for widget in widgets:
            widget.renderable = widget.render_title()

    def on_mount(self):
        """Initiate rendering in the background."""
        self.run_worker(self.render_all_properties, thread=True)


class PropertyValues(Widget):
    """Container for property values."""

    MAX_COLUMN_COUNT = 6

    DEFAULT_CSS = """
    PropertyValues {
        layout: grid;
        grid-size: 6;
        height: auto;
        max-width: 85%;
    }
    """

    def on_mount(self):
        """Adjust column count based on children of this node."""
        children_count = len(self.children)
        self.styles.grid_size_columns = max(
            min(
                children_count,
                self.MAX_COLUMN_COUNT,
            ),
            1,
        )


class PropertyRow(Widget, can_focus=False, inherit_bindings=False):
    """A container with horizontal layout and no scrollbars."""

    DEFAULT_CSS = """
    PropertyRow {
        width: 1fr;
        height: auto;
        layout: horizontal;
    }
    """


class LiteralPropertyValue(Widget, can_focus=True, inherit_bindings=False):
    """
    Literal value of a property.

    Supports provenance.
    """

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding('p', 'provenance', 'Provenan©e'),
    ]

    DEFAULT_CSS = """
    LiteralPropertyValue {
        width: auto;
        height: auto;
    }

    LiteralPropertyValue:hover {
        background: $boost;
    }

    LiteralPropertyValue:focus {
        background: darkslateblue;
    }
    """

    renderable: str | None = reactive[str | None](   # noqa: WPS465
        None,
        init=False,
        layout=True,
    )

    def __init__(
        self,
        property_value: Literal,
        subject: NotLiteralNode,
        property_iri: NotLiteralNode,
    ):
        """Initialize parameters for rendering, navigation, & provenance."""
        self.property_value = property_value
        self.subject = subject
        self.property_iri = property_iri
        super().__init__()
        self.renderable = Text(   # noqa: WPS601
            str(property_value),
            style='#696969',
        )

    @property
    def iri(self):
        """Return the property IRI for compatibility."""
        return self.property_value

    def render(self) -> RenderResult:
        """Render title of the node."""
        return self.renderable

    def action_provenance(self):
        """Navigate to provenance for the property value."""
        self.app.action_goto(
            TripleURIRef.from_triple(
                subject=self.subject,
                predicate=self.property_iri,
                object_=self.property_value,
            ),
        )


class Title(Static):
    """Iolanta page title."""

    DEFAULT_CSS = """
    Title {
        padding: 1;
        background: darkslateblue;
        color: white;
        text-style: bold;
    }
    """


class ContentArea(VerticalScroll):
    """Description of the IRI."""

    DEFAULT_CSS = """
    Content {
        layout: vertical;
        height: auto;
        max-height: 100%;
    }

    #description {
        padding: 1;
    }

    #properties {
        padding: 1;
    }

    /* FIXME: This one does not work */
    DataTable .datatable--header {
        background: purple;
        color: red;
    }
    """

    def compose(self) -> ComposeResult:
        """Render tabs."""
        with TabbedContent():
            for label, tab_content in self.tabs.items():
                if tab_content is None:
                    raise ValueError(
                        f'Tab `{label}` is `None` and is not renderable.',
                    )

                with TabPane(label):
                    yield tab_content



from textual.widgets import Static
from textual.worker import Worker, WorkerState

from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES
from iolanta.widgets.mixin import IolantaWidgetMixin


class PageTitle(IolantaWidgetMixin, Static):
    """Iolanta page title."""

    DEFAULT_CSS = """
    PageTitle {
        padding: 1;
        background: darkslateblue;
        color: white;
        text-style: bold;
    }
    """

    def __init__(self, iri: NotLiteralNode, extra=None) -> None:
        """Initialize."""
        self.iri = iri
        self.extra = extra
        super().__init__(iri)

    def construct_title(self):
        """Render the title via Iolanta in a thread."""
        return self.iolanta.render(
            self.iri,
            as_datatype=DATATYPES.title,
        )

    def on_mount(self):
        """Initialize rendering of a title."""
        self.run_worker(self.construct_title, thread=True)

    def on_worker_state_changed(   # noqa: WPS210
        self,
        event: Worker.StateChanged,
    ):
        """Render title when generated."""
        match event.state:
            case WorkerState.SUCCESS:
                title = event.worker.result
                if self.extra:
                    title = f'{title} {self.extra}'
                self.update(title)

            case WorkerState.ERROR:
                raise ValueError(event)



import enum
import hashlib
import re
import textwrap
from urllib import parse as urllib_parse

from documented import Documented
from pydantic import AnyUrl, BaseModel
from rdflib import BNode, Literal, URIRef

from iolanta.models import NotLiteralNode  # noqa: WPS202


def escape_label(label: str) -> str:
    """Escape a label and return it wrapped in appropriate quotes.

    Returns the label with URLs stripped, quotes escaped, and wrapped in quotes.
    Uses single quotes if label contains double quotes to avoid escaping issues.

    Escapes quotes in labels that will be wrapped in quotes in Mermaid syntax.
    """
    # Remove https://, http://, and www. prefixes to prevent markdown link parsing
    safe_label = (
        label.replace("https://", "").replace("http://", "").replace("www.", "")
    )
    # Use single quotes if label contains double quotes to avoid escaping issues
    use_single = '"' in safe_label
    quote_char = "'" if use_single else '"'
    # Escape the quote character that will be used for wrapping
    if use_single:
        escaped_label = safe_label.replace("'", r"\'")
    else:
        escaped_label = safe_label.replace('"', r"\"")
    return f"{quote_char}{escaped_label}{quote_char}"


class Direction(enum.StrEnum):
    """Mermaid diagram direction."""

    TB = "TB"
    LR = "LR"


class MermaidScalar(Documented, BaseModel, arbitrary_types_allowed=True):
    """Base class for Mermaid scalar elements (nodes and edges)."""

    @property
    def id(self) -> str:
        """Get the unique identifier for this Mermaid element."""
        raise NotImplementedError()


class MermaidURINode(MermaidScalar, frozen=True):
    """
    {self.id}{self.maybe_title}
    click {self.id} "{self.url}"
    """

    uri: URIRef
    url: AnyUrl
    title: str = ""

    @property
    def maybe_title(self):
        if not self.title:
            return ""
        quoted_title = escape_label(self.title)
        return f"({quoted_title})"

    @property
    def id(self):
        return re.sub(
            r"[:\/\.#()?=&+]", "_", urllib_parse.unquote(str(self.url)).strip("/")
        )


class MermaidLiteral(MermaidScalar, frozen=True):
    """{self.id}[["{self.title}"]]"""

    literal: Literal

    @property
    def title(self) -> str:
        raw_title = str(self.literal) or "EMPTY"
        # Replace quotes with safer characters for Mermaid
        return raw_title.replace('"', '"').replace("'", "'")

    @property
    def id(self) -> str:
        # Use the lexical form of the literal, not rdflib's .value (which may be empty for typed literals),
        # to ensure different texts get distinct node IDs in Mermaid.
        value_hash = hashlib.md5(str(self.literal).encode()).hexdigest()
        return f"Literal-{value_hash}"


class MermaidBlankNode(MermaidScalar):
    """{self.id}({self.escaped_title})"""

    node: BNode
    title: str

    @property
    def id(self) -> str:
        return self.node.replace("_:", "")

    @property
    def escaped_title(self) -> str:
        """Escape the title to prevent Mermaid parsing issues."""
        return escape_label(self.title)


class MermaidEdge(MermaidScalar):
    """
    {self.source.id} --- {self.id}([{self.escaped_title}])--> {self.target.id}
    click {self.id} "{self.predicate}"
    class {self.id} predicate
    """

    source: "MermaidURINode | MermaidBlankNode | MermaidSubgraph"
    target: "MermaidURINode | MermaidLiteral | MermaidBlankNode | MermaidSubgraph"
    predicate: URIRef
    title: str

    @property
    def id(self) -> str:
        return hashlib.md5(
            f"{self.source.id}{self.predicate}{self.target.id}".encode()  # noqa: WPS237
        ).hexdigest()

    @property
    def nodes(self):
        return [self.source, self.target]

    @property
    def escaped_title(self) -> str:
        # Escape URLs to prevent Mermaid from interpreting them as markdown links
        return escape_label(self.title)


class MermaidSubgraph(Documented, BaseModel, arbitrary_types_allowed=True, frozen=True):
    """
    subgraph {self.id}[{self.escaped_title}]
      direction {self.direction}
      {self.formatted_body}
    end
    """

    children: list[MermaidScalar]
    uri: NotLiteralNode
    title: str
    direction: Direction = Direction.LR

    @property
    def id(self):
        uri_hash = hashlib.md5(str(self.uri).encode()).hexdigest()
        return f"subgraph_{uri_hash}"

    @property
    def escaped_title(self) -> str:
        """Escape the subgraph title to prevent markdown link parsing."""
        return escape_label(self.title)

    @property
    def formatted_body(self):
        return textwrap.indent(
            "\n".join(map(str, self.children)),
            prefix="  ",
        )


class Diagram(Documented, BaseModel):
    """
    graph {self.direction}
    {self.formatted_body}
      classDef predicate fill:transparent,stroke:transparent,stroke-width:0px;
      {self.formatted_tail}
    """

    children: list[MermaidScalar | MermaidSubgraph]
    direction: Direction = Direction.LR
    tail: str | None = None

    @property
    def formatted_tail(self) -> str:
        return self.tail or ""

    @property
    def formatted_body(self):
        return textwrap.indent(
            "\n".join(map(str, self.children)),
            prefix="  ",
        )



import functools
from dataclasses import dataclass, field, replace
from pathlib import Path
from typing import (  # noqa: WPS235
    Annotated,
    Any,
    Iterable,
    List,
    Optional,
    Protocol,
    Set,
    Type,
)

import loguru
import yaml_ld
from pyparsing import ParseException
from rdflib import ConjunctiveGraph, Graph, Literal, URIRef
from rdflib.namespace import NamespaceManager
from rdflib.plugins.parsers.notation3 import BadSyntax
from rdflib.plugins.sparql.processor import SPARQLResult
from rdflib.term import Node
from yaml_ld.document_loaders.content_types import ParserNotFound
from yaml_ld.errors import YAMLLDError

from iolanta import entry_points, namespaces
from iolanta.conversions import path_to_iri
from iolanta.errors import UnresolvedIRI
from iolanta.facets.errors import FacetError
from iolanta.facets.locator import FacetFinder
from iolanta.models import ComputedQName, LDContext, NotLiteralNode
from iolanta.node_to_qname import node_to_qname
from iolanta.parse_quads import parse_quads
from iolanta.plugin import Plugin
from iolanta.query_result import (
    QueryResult,
    SPARQLParseException,
    SPARQLQueryArgument,
    format_query_bindings,
)
from iolanta.resolvers.base import Resolver
from iolanta.resolvers.dispatch import SchemeDispatchResolver
from iolanta.resolvers.pypi import PyPIResolver
from iolanta.resolvers.python_import import PythonImportResolver
from iolanta.sparqlspace.processor import normalize_term  # noqa: WPS201


class LoggerProtocol(Protocol):
    """Abstract Logger interface that unites `loguru` & standard `logging`."""

    def info(self, message: str, *args: Any, **kwargs: Any) -> None:  # noqa: WPS110
        """Log an INFO message."""

    def error(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log an ERROR message."""

    def warning(self, message: str, *args: Any, **kwargs: Any) -> None:
        """Log a WARNING message."""


def _create_default_graph():
    return ConjunctiveGraph(identifier=namespaces.LOCAL.term("_inference"))


@dataclass
class Iolanta:  # noqa: WPS214, WPS338
    """Iolanta is a Semantic web browser."""

    language: Literal = Literal("en")

    project_root: Annotated[
        Path | None,
        (
            "File or directory the contents of which "
            "Iolanta will automatically load into the graph."
        ),
    ] = None
    graph: ConjunctiveGraph = field(default_factory=_create_default_graph)
    force_plugins: List[Type[Plugin]] = field(default_factory=list)

    facet_resolver: Resolver = field(
        default_factory=functools.partial(
            SchemeDispatchResolver,
            python=PythonImportResolver,
            pkg=PyPIResolver,
        ),
    )

    logger: LoggerProtocol = loguru.logger

    could_not_retrieve_nodes: Set[Node] = field(
        default_factory=set,
        init=False,
    )

    _facet_inference_needed: bool = field(
        default=False,
        init=False,
    )

    @property
    def plugin_classes(self) -> List[Type[Plugin]]:
        """Installed Iolanta plugins."""
        return self.force_plugins or entry_points.plugins("iolanta.plugins")

    @functools.cached_property
    def plugins(self) -> List[Plugin]:
        """Construct a list of installed plugin instances."""
        return [plugin_class(iolanta=self) for plugin_class in self.plugin_classes]

    def _run_facet_inference(self):
        """Run inference queries from facet inference directories."""
        for facet_class in self.facet_classes:
            inference_path = facet_class.inference_path

            if inference_path is not None:
                facet_name = facet_class.__name__.lower()
                self._run_inference_from_directory(
                    inference_path,
                    graph_prefix=f"facet-{facet_name}",
                )

    def _run_inference_from_directory(  # noqa: WPS210, WPS231
        self,
        inference_dir: Path,
        graph_prefix: str = "inference",
    ):
        """
        Run inference queries from a given inference directory.

        For each SPARQL file in the inference directory:
        1. Truncate the named graph `local:{graph_prefix}-{filename}`
        2. Execute the CONSTRUCT query
        3. Insert the resulting triples into that graph
        """
        if not inference_dir.exists():
            return

        inference_files = sorted(inference_dir.glob("*.sparql"))

        for inference_file in inference_files:
            filename = inference_file.stem  # filename without .sparql extension
            inference_graph = URIRef(f"{graph_prefix}:{filename}")

            # Truncate the inference graph
            context = self.graph.get_context(inference_graph)
            context.remove((None, None, None))

            # Read and execute the CONSTRUCT query
            query_text = inference_file.read_text()
            try:
                query_result = self.graph.query(query_text)  # noqa: WPS110
            except (SyntaxError, ParseException) as err:
                self.logger.error(
                    "Invalid SPARQL syntax in inference query {filename}: {error}",
                    filename=filename,
                    error=err,
                )
                raise SPARQLParseException(
                    error=err,
                    query=query_text,
                ) from err

            # CONSTRUCT queries return a SPARQLResult with a graph attribute
            result_graph = (
                query_result.get("graph")
                if isinstance(query_result, dict)
                else query_result.graph
            )
            if result_graph is None:
                raise ValueError(
                    f"CONSTRUCT query {filename} returned None result_graph. "
                    f"query_result type: {type(query_result)}, "
                    f"query_result: {query_result}"
                )

            inferred_quads = [
                (s, p, o, inference_graph)  # noqa: WPS111
                for s, p, o in result_graph  # noqa: WPS111
            ]
            self.logger.info(f"Inference {filename}: generated quads: {inferred_quads}")

            if inferred_quads:
                self.graph.addN(inferred_quads)  # noqa: WPS220
                self.logger.info(  # noqa: WPS220
                    "Inference {filename}: added {count} triples",
                    filename=filename,
                    count=len(inferred_quads),
                )

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """Run a SPARQL `SELECT`, `CONSTRUCT`, or `ASK` query."""
        try:
            sparql_result: SPARQLResult = self.graph.query(
                query_text,
                processor="sparqlspace",
                initBindings=kwargs,
            )
        except SyntaxError as err:
            raise SPARQLParseException(
                error=err,
                query=query_text,
            ) from err

        # Run facet-specific inference if needed
        if self._facet_inference_needed:
            self._run_facet_inference()
            self._facet_inference_needed = False

        if sparql_result.askAnswer is not None:
            return sparql_result.askAnswer

        if sparql_result.graph is not None:
            graph: Graph = sparql_result.graph
            for prefix, namespace in self.graph.namespaces():
                graph.bind(prefix, namespace)

            return graph

        return format_query_bindings(sparql_result.bindings)

    def reset(self):
        """Reset Iolanta graph."""
        self.graph = _create_default_graph()  # noqa: WPS601
        self._facet_inference_needed = False
        self.__post_init__()

    def add(  # noqa: C901, WPS231, WPS210, WPS213
        self,
        source: Path,
        context: Optional[LDContext] = None,
        graph_iri: Optional[URIRef] = None,
    ) -> "Iolanta":
        """Parse & load information from given URL into the graph."""
        self.logger.info(f"Adding to graph: {source}")

        if not isinstance(source, Path):
            source = Path(source)

        for source_file in list(source.rglob("*")) or [source]:
            if source_file.is_dir():
                continue

            try:  # noqa: WPS225
                ld_rdf = yaml_ld.to_rdf(source_file)
            except ConnectionError as name_resolution_error:
                self.logger.warning(
                    "%s | name resolution error: %s",
                    source_file,
                    str(name_resolution_error),
                )
                continue
            except ParserNotFound as parser_not_found:
                self.logger.error(f"{source} | {parser_not_found}")
                continue
            except (YAMLLDError, BadSyntax) as parse_error:
                self.logger.warning("%s | parse error: %s", source_file, parse_error)
                file_iri = path_to_iri(source_file)
                self.graph.addN(
                    [
                        (
                            file_iri,
                            namespaces.IOLANTA["parse-error"],
                            Literal(str(parse_error)),
                            namespaces.META,
                        ),
                    ]
                )
                continue
            except ValueError as value_error:
                self.logger.error(f"{source} | {value_error}")
                continue

            self.logger.info(f"{source_file} is loaded.")

            graph = path_to_iri(source_file)
            try:
                quads = list(
                    parse_quads(
                        quads_document=ld_rdf,
                        graph=graph,
                        blank_node_prefix=str(source_file),
                    ),
                )
            except UnresolvedIRI as err:
                raise replace(
                    err,
                    context=None,
                    iri=graph,
                )

            if not quads:
                self.logger.info(f"{source_file} | No data found")
                continue

            self.graph.addN(quads)
            self._facet_inference_needed = True

        return self

    def infer(self, closure_class=None) -> "Iolanta":
        """Apply inference."""
        return self

    def bind_namespaces(self):  # noqa: WPS213
        """Bind namespaces."""
        self.graph.namespace_manager = NamespaceManager(
            self.graph,
            bind_namespaces="none",
        )
        self.graph.bind(prefix="local", namespace=namespaces.LOCAL)
        self.graph.bind(prefix="iolanta", namespace=namespaces.IOLANTA)
        self.graph.bind(prefix="rdf", namespace=namespaces.RDF)
        self.graph.bind(prefix="rdfs", namespace=namespaces.RDFS)
        self.graph.bind(prefix="owl", namespace=namespaces.OWL)
        self.graph.bind(prefix="foaf", namespace=namespaces.FOAF)
        self.graph.bind(prefix="schema", namespace=namespaces.SDO)
        self.graph.bind(prefix="vann", namespace=namespaces.VANN)
        self.graph.bind(prefix="np", namespace=namespaces.NP)
        self.graph.bind(prefix="dcterms", namespace=namespaces.DCTERMS)
        self.graph.bind(prefix="rdfg", namespace=namespaces.RDFG)

    @functools.cached_property
    def context_paths(self) -> Iterable[Path]:
        """Compile list of context files."""
        directory = Path(__file__).parent / "data"

        yield directory / "context.yaml"

        for plugin in self.plugins:
            if path := plugin.context_path:  # noqa: WPS332
                yield path

    def add_files_from_plugins(self):
        """Load files from plugins."""
        for plugin in self.plugins:
            try:
                self.add(plugin.data_files)
            except Exception as error:
                self.logger.error(
                    f"Cannot load {plugin} plugin data files: {error}",
                )

    @property
    def facet_classes(self):
        """Get all registered facet classes."""
        return entry_points.plugins("iolanta.facets")

    def add_files_from_facets(self):
        """Add files from all registered facets to the graph."""
        for facet_class in self.facet_classes:
            try:
                self.add(facet_class.META)
            except AttributeError:
                pass  # noqa: WPS420

    def __post_init__(self):
        """Initialize after instance creation."""
        self.bind_namespaces()
        self.add_files_from_plugins()
        self.add_files_from_facets()
        if self.project_root:
            self.add(self.project_root)

    def render(
        self,
        node: Node,
        as_datatype: NotLiteralNode,
    ) -> Any:
        """Find an Iolanta facet for a node and render it."""
        node = normalize_term(node)
        if not as_datatype:
            raise ValueError(
                f"Please provide the datatype to render {node} as.",
            )

        if isinstance(as_datatype, list):
            raise NotImplementedError("Got a list for as_datatype :(")

        found = FacetFinder(
            iolanta=self,
            node=node,
            as_datatype=as_datatype,
        ).facet_and_output_datatype

        facet_class = self.facet_resolver.resolve(found["facet"])

        facet = facet_class(
            this=node,
            iolanta=self,
            as_datatype=found["output_datatype"],
        )

        try:
            return facet.show()

        except Exception as err:
            raise FacetError(
                node=node,
                facet_iri=found["facet"],
                error=err,
            ) from err

    def node_as_qname(self, node: Node):
        """Render node as a QName if possible."""
        qname = node_to_qname(node, self.graph)
        return (
            f"{qname.namespace_name}:{qname.term}"
            if isinstance(
                qname,
                ComputedQName,
            )
            else node
        )



from pathlib import Path

import funcy
from rich.markdown import Markdown
from textual.containers import Vertical
from yarl import URL

from iolanta import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.widgets.description import Description

TEXT = """
**😕 Iolanta is unable to visualize this resource**

* The reference type ({reference_type}) might be incorrect;
* The URI might be incorrect;
* Or, no edges might exist which involve it;
* Or maybe Iolanta does not know of such edges.
{content}
{subgraphs}
**What can you do?**

* If you feel this might indicate a bug 🐛, please do let us know at GitHub
issues: https://github.com/iolanta.tech/iolanta/issues
"""

PARSE_ERROR_TEMPLATE = """
**Parse error**

```
{message}
```
"""

CONTENT_TEMPLATE = """
**File content**

```{type}
{content}
```
"""

SUBGRAPHS_TEMPLATE = """
**Subgraphs**

{formatted_subgraphs}
"""


class TextualNoFacetFound(Facet):
    """Facet to handle the case when no facet is found."""

    @property
    def raw_content(self):
        """Content of the file, if applicable."""
        url = URL(self.this)
        if url.scheme != "file":
            return None

        path = Path(url.path)
        if not path.is_relative_to(self.iolanta.project_root):
            return None

        if not path.exists():
            return None

        if path.is_dir():
            return None

        parse_error_msg = self._parse_error_message()
        file_content = path.read_text()
        content = CONTENT_TEMPLATE.format(
            content=file_content,
            type={
                ".yamlld": "yaml",
                ".jsonld": "json",
                ".ttl": "turtle",
            }.get(path.suffix, ""),
        )
        if parse_error_msg:
            content = PARSE_ERROR_TEMPLATE.format(message=parse_error_msg) + content
        return content

    def _parse_error_message(self) -> str | None:
        """Return parse error message for this node if it failed to load."""
        rows = self.query(
            """
            SELECT ?msg WHERE {
                GRAPH <iolanta://_meta> { $this iolanta:parse-error ?msg }
            }
            """,
            this=self.this,
        )
        if row := funcy.first(rows):
            return str(row["msg"])
        return None

    @property
    def subgraphs_description(self) -> str:
        """Return a formatted description of subgraphs, if any exist."""
        rows = self.query(
            "SELECT ?subgraph WHERE { $this iolanta:has-sub-graph ?subgraph }",
            this=self.this,
        )
        subgraphs = funcy.lpluck("subgraph", rows)
        if subgraphs:
            return SUBGRAPHS_TEMPLATE.format(
                formatted_subgraphs="\n".join(
                    [f"- {subgraph}" for subgraph in subgraphs]
                ),
            )

        return ""

    def show(self):
        """Compose the page."""
        return Vertical(
            PageTitle(self.this),
            Description(
                Markdown(
                    TEXT.format(
                        content=self.raw_content or "",
                        subgraphs=self.subgraphs_description or "",
                        reference_type=type(self.this).__name__,
                    ),
                ),
            ),
        )



from textual.app import ComposeResult
from textual.widgets import Placeholder, Static


class Home(Placeholder):
    """Default home page."""

    def compose(self) -> ComposeResult:
        """
        Show default content.

        TODO: Implement a set of default links, search or something here.
        """
        yield Static('Welcome to Iolanta! This is a placeholder page.')



from collections import defaultdict
from pathlib import Path
from typing import Iterable, NamedTuple

import funcy
from rdflib import Literal, Node
from textual.containers import Vertical
from textual.coordinate import Coordinate
from textual.widgets import DataTable

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES
from iolanta.widgets.mixin import IolantaWidgetMixin


class GraphRow(NamedTuple):
    """A row in the graphs table."""

    graph: NotLiteralNode
    count: int


class GraphsTable(IolantaWidgetMixin, DataTable):
    """Render graphs as a table with graph URI and triple count."""

    BINDINGS = [
        ('enter', 'goto', 'Goto'),
    ]

    def __init__(self, rows: Iterable[GraphRow]):
        """Construct."""
        super().__init__(show_header=True, cell_padding=1)
        self.graph_rows = list(rows)

    def render_human_readable_cells(self):
        """Replace the cells with their human readable titles."""
        terms_and_coordinates = sorted(
            self.node_to_coordinates.items(),
            key=lambda node_and_coordinates_pair: len(
                node_and_coordinates_pair[1],
            ),
            reverse=True,
        )

        for term, coordinates in terms_and_coordinates:
            title = str(self.iolanta.render(term, as_datatype=DATATYPES.title))
            for coordinate in coordinates:
                self.app.call_from_thread(
                    self.update_cell_at,
                    coordinate,
                    value=title,
                    update_width=False,
                )

    @funcy.cached_property
    @funcy.post_processing(dict)
    def coordinate_to_node(self):
        """Return a mapping of coordinates to their corresponding nodes."""
        for row_number, (graph, _count) in enumerate(self.graph_rows):
            yield Coordinate(row_number, 0), graph

    @funcy.cached_property
    def node_to_coordinates(self) -> defaultdict[Node, list[Coordinate]]:
        """Map node to coordinates where it appears."""
        node_to_coordinate = [
            (node, coordinate)
            for coordinate, node in self.coordinate_to_node.items()
        ]
        return funcy.group_values(node_to_coordinate)

    def format_as_loading(self, node: Node) -> str:
        """Intermediate version of a value while it is loading."""
        if isinstance(node, Literal):
            node_text = f'⌛ {node}'
        else:
            node_text = self.iolanta.node_as_qname(node)
            node_text = f'⌛ {node_text}'

        return node_text

    def on_mount(self):
        """Fill the table and start rendering."""
        self.add_columns('Graph', 'Triples Count')
        self.cell_padding = 1

        for graph, count in self.graph_rows:
            self.add_row(
                self.format_as_loading(graph),
                str(count),
            )

        self.run_worker(
            self.render_human_readable_cells,
            thread=True,
        )

    def action_goto(self):
        """Navigate to the selected graph."""
        if self.cursor_coordinate:
            node = self.coordinate_to_node.get(self.cursor_coordinate)
            if node is not None:
                self.app.action_goto(node)


class GraphsBody(Vertical):
    """Container for graphs table."""

    DEFAULT_CSS = """
    GraphsBody {
        height: auto;
        max-height: 100%;
    }
    """


class Graphs(Facet):
    """Render named graphs as a table."""

    META = Path(__file__).parent / 'data' / 'textual_graphs.yamlld'

    def show(self):
        """Construct the table."""
        rows = self.stored_query('graphs.sparql')

        return GraphsBody(
            PageTitle(self.this),
            GraphsTable(
                [
                    GraphRow(
                        graph=row['graph'],
                        count=int(row['count'].value),
                    )
                    for row in rows
                ],
            ),
        )



import re
from types import MappingProxyType

from rdflib import URIRef

# All entries are (pattern, replacement) strings; pattern may be regex with $ for exact match.
REDIRECTS = MappingProxyType(
    {
        # FIXME This is presently hardcoded; we need to
        #   - either find a way to resolve these URLs automatically,
        #   - or create a repository of those redirects online.
        r"http://purl\.org/vocab/vann/$": "https://vocab.org/vann/vann-vocab-20100607.rdf",
        r"https://purl\.org/dc/elements/1\.1/$": "https://purl.org/dc/terms/",
        r"http://www\.w3\.org/1999/02/22-rdf-syntax-ns#$": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        r"http://www\.w3\.org/2000/01/rdf-schema#$": "http://www.w3.org/2000/01/rdf-schema#",
        r"http://www\.w3\.org/2002/07/owl#$": "http://www.w3.org/2002/07/owl#",
        # Add # fragment to OWL and RDFS when URI has no fragment (pattern $ = no fragment)
        # (fixes bug reported at https://stackoverflow.com/q/78934864/1245471)
        r"http://www\.w3\.org/2002/07/owl$": "http://www.w3.org/2002/07/owl#",
        r"http://www\.w3\.org/2000/01/rdf-schema$": "http://www.w3.org/2000/01/rdf-schema#",
        # Redirect FOAF namespace to GitHub mirror
        r"https?://xmlns\.com/foaf/0\.1/.+": "https://raw.githubusercontent.com/foaf/foaf/refs/heads/master/xmlns.com/htdocs/foaf/0.1/index.rdf",
        r"https://www\.nanopub\.org/nschema$": "https://www.nanopub.net/nschema#",
        r"https://nanopub\.org/nschema$": "https://nanopub.net/nschema#",
        # Convert lexvo.org/id URLs to lexvo.org/data URLs
        r"http://lexvo\.org/id/(.+)": r"http://lexvo.org/data/\1",
        r"https://lexvo\.org/id/(.+)": r"http://lexvo.org/data/\1",
        r"https://www\.lexinfo\.net/(.+)": r"http://www.lexinfo.net/\1",
        # Convert Wikidata https:// to http:// (Wikidata JSON-LD uses http:// URIs)
        r"https://www\.wikidata\.org/entity/(.+)": r"http://www.wikidata.org/entity/\1",
    }
)


def apply_redirect(source: URIRef) -> URIRef:  # noqa: WPS210
    """
    Rewrite the URL using regex patterns and group substitutions.

    For each pattern in REDIRECTS:
    - If the pattern matches the source URI
    - Replace the source with the destination, substituting any regex groups
    """
    source_str = str(source)

    for pattern, destination in REDIRECTS.items():
        pattern_str = str(pattern)
        destination_str = str(destination)

        if re.match(pattern_str, source_str):
            # Replace any group references in the destination (e.g. \1)
            redirected_uri = re.sub(
                pattern_str,
                destination_str,
                source_str,
            )
            return URIRef(redirected_uri)

    return source



from typing import Iterable

from textual.containers import Horizontal

from iolanta.facets.textual_nanopublication.term_widget import TermWidget
from iolanta.widgets.mixin import IolantaWidgetMixin


class TermList(IolantaWidgetMixin, Horizontal):
    """Display a sequence of terms."""

    DEFAULT_CSS = """
    TermList {
        padding: 1 2;
        height: auto;
    }
    """

    def __init__(self, terms: Iterable[TermWidget]):
        """Initialize."""
        self.terms = terms
        super().__init__()

    def compose(self):
        """Fill in the term stubs."""
        yield from self.terms

    def on_mount(self):
        """Initialize terms rendering."""
        self.run_worker(self.render_terms, thread=True)

    def render_terms(self):
        """Render terms."""
        child: TermWidget
        for child in self.children:
            child.renderable = self.iolanta.render(
                child.uri,
                as_datatype=child.as_datatype,
            )



import csv
import io
from pathlib import Path
from typing import cast

from rdflib import Graph, Literal

from iolanta.facets.cli.base import Renderable, RichFacet
from iolanta.facets.errors import NotALiteral

META = Path(__file__).parent / 'data' / 'query_result.yamlld'


class ConstructResultCsvFacet(RichFacet):
    """Render CONSTRUCT query results as CSV."""

    META = META

    def show(self) -> Renderable:
        """Render Graph (CONSTRUCT result) as CSV."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(node=self.this)

        graph = cast(Graph, self.this.value)

        if not graph:
            return ""

        output = io.StringIO()
        writer = csv.writer(output)
        writer.writerow(('subject', 'predicate', 'object'))
        writer.writerows(graph)

        return output.getvalue()



import funcy
from rdflib import URIRef

from iolanta.facets.facet import Facet
from iolanta.namespaces import DATATYPES


class WikibaseStatementTitle(Facet[str]):
    """Calculate title for a Wikibase Statement."""

    def show(self) -> str:
        """Render the title."""
        rows = self.stored_query(
            'statement-title.sparql',
            statement=self.this,
            language=self.language,
        )

        row = funcy.first(rows)
        if not row:
            return self.render(
                self.this,
                as_datatype=URIRef('https://iolanta.tech/qname'),
            )

        return self.render(
            row['entity'],
            as_datatype=DATATYPES.title,
        )



from collections import defaultdict
from typing import Iterable

import funcy
from rdflib import Literal, Node
from rich.style import Style
from rich.text import Text
from textual.containers import Vertical
from textual.coordinate import Coordinate
from textual.widgets import DataTable

from iolanta import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DATATYPES
from iolanta.widgets.mixin import IolantaWidgetMixin

PAIRS_QUERY = """
SELECT ?subject ?object WHERE {
    ?subject $this ?object .
} ORDER BY ?subject ?object
"""


class PairsTable(IolantaWidgetMixin, DataTable):
    """Render subject → object properties as a table with two columns."""

    BINDINGS = [
        ('enter', 'goto', 'Goto'),
    ]

    def __init__(self, pairs: Iterable[tuple[NotLiteralNode, Node]]):
        """Construct."""
        super().__init__(show_header=False, cell_padding=1)
        self.pairs = list(pairs)

    def render_human_readable_cells(self):
        """Replace the cells with their human readable titles."""
        terms_and_coordinates = sorted(
            self.node_to_coordinates.items(),
            key=lambda node_and_coordinates_pair: len(
                node_and_coordinates_pair[1],
            ),
            reverse=True,
        )

        for term, coordinates in terms_and_coordinates:
            title = self.iolanta.render(term, as_datatype=DATATYPES.title)
            for coordinate in coordinates:
                self.app.call_from_thread(
                    self.update_cell_at,
                    coordinate,
                    value=Text(title, no_wrap=False),
                    update_width=False,
                )

    @funcy.cached_property
    @funcy.post_processing(dict)
    def coordinate_to_node(self):
        """Return a mapping of coordinates to their corresponding nodes."""
        for row_number, (subject, object_term) in enumerate(self.pairs):
            yield Coordinate(row_number, 0), subject
            yield Coordinate(row_number, 1), object_term

    @funcy.cached_property
    def node_to_coordinates(self) -> defaultdict[Node, list[Coordinate]]:
        """Map node to coordinates where it appears."""
        node_to_coordinate = [
            (node, coordinate)
            for coordinate, node in self.coordinate_to_node.items()
        ]
        return funcy.group_values(node_to_coordinate)

    def format_as_loading(self, node: Node) -> Text:
        """Intermediate version of a value while it is loading."""
        if isinstance(node, Literal):
            node_text = f'⌛ {node}'
        else:
            node_text = self.iolanta.node_as_qname(node)
            node_text = f'⌛ {node_text}'

        return Text(
            node_text,
            style=Style(dim=True),
            no_wrap=False,
        )

    def on_mount(self):
        """Fill the table and start rendering."""
        self.add_columns('Subject', 'Object')
        self.cell_padding = 1

        for subject, object_term in self.pairs:
            self.add_row(
                self.format_as_loading(subject),
                self.format_as_loading(object_term),
            )

        self.run_worker(
            self.render_human_readable_cells,
            thread=True,
        )

    def action_goto(self):
        """Navigate to the selected node."""
        if self.cursor_coordinate:
            node = self.coordinate_to_node.get(self.cursor_coordinate)
            if node is not None:
                self.app.action_goto(node)


class TextualPropertyPairsTableFacet(Facet):
    """Render a table of subject → object pairs for a property."""

    def show(self):
        """Construct the table."""
        rows = self.query(
            PAIRS_QUERY,
            this=self.this,
        )

        return Vertical(
            PageTitle(
                self.this,
                extra='— Subjects & Objects connected by this property',
            ),
            PairsTable(
                [
                    (row['subject'], row['object'])
                    for row in rows
                ],
            ),
        )



import json
from pathlib import Path
from typing import cast

from rdflib import Graph, Literal

from iolanta.facets.cli.base import Renderable, RichFacet
from iolanta.facets.errors import NotALiteral

META = Path(__file__).parent / 'data' / 'query_result.yamlld'


class ConstructResultJsonFacet(RichFacet):
    """Render CONSTRUCT query results as JSON."""

    META = META

    def show(self) -> Renderable:
        """Render Graph (CONSTRUCT result) as JSON."""
        if not isinstance(self.this, Literal):
            raise NotALiteral(node=self.this)

        graph = cast(Graph, self.this.value)
        fieldnames = ('subject', 'predicate', 'object')
        return json.dumps(
            [
                dict(zip(fieldnames, triple))
                for triple in graph
            ],
            indent=2,
            default=str,
        )



import functools
from pathlib import Path
from xml.dom import minidom  # noqa: S408

import funcy
from rdflib.term import Literal, Node
from rich.syntax import Syntax
from textual.containers import Vertical, VerticalScroll
from textual.widget import Widget
from textual.widgets import Static

from iolanta.facets.facet import Facet
from iolanta.facets.page_title import PageTitle
from iolanta.facets.textual_default.widgets import (
    LiteralPropertyValue,
    PropertiesContainer,
    PropertyName,
    PropertyRow,
    PropertyValue,
    PropertyValues,
)
from iolanta.models import NotLiteralNode
from iolanta.namespaces import DC, RDFS, SDO


class Description(Static):
    """Static widget with padding for descriptions."""

    DEFAULT_CSS = """
    Description {
        padding: 1;
    }
    """


class TextualDefaultFacet(Facet[Widget]):   # noqa: WPS214
    """Default rendering engine."""

    query_file_name = 'properties.sparql'
    properties_on_the_right = False

    @functools.cached_property
    def grouped_properties(self) -> dict[NotLiteralNode, list[Node]]:
        """Properties of current node & their values."""
        property_rows = self.stored_query(
            self.query_file_name,
            this=self.this,
        )

        property_pairs = [
            (row['property'], row['object'])
            for row in property_rows
        ]

        property_pairs = [
            (property_iri, object_node)
            for property_iri, object_node in property_pairs
            if (
                not isinstance(object_node, Literal)
                or not (language := object_node.language)  # noqa: W503
                or (language == self.iolanta.language)     # noqa: W503
            )
        ]

        return funcy.group_values(property_pairs)

    @functools.cached_property
    def rows(self):
        """Generate rows for the properties table."""
        for property_iri, property_values in self.grouped_properties.items():
            property_name = PropertyName(
                iri=property_iri,
                qname=self.iolanta.node_as_qname(property_iri),
            )

            property_values = [
                LiteralPropertyValue(
                    property_value=property_value,
                    subject=self.this,
                    property_iri=property_iri,
                ) if isinstance(property_value, Literal) else PropertyValue(
                    property_value=property_value,
                    subject=self.this,
                    property_iri=property_iri,
                    property_qname=self.iolanta.node_as_qname(property_iri),
                )
                for property_value in property_values
            ]

            if self.properties_on_the_right:
                # Property name on the right (last column), values on the left (first column)
                yield PropertyRow(
                    PropertyValues(*property_values),
                    property_name,
                )
            else:
                # Property name on the left (first column), values on the right (last column)
                yield PropertyRow(
                    property_name,
                    PropertyValues(*property_values),
                )

    @functools.cached_property
    def description(self) -> str | Syntax | None:
        """
        Candidates for description.

        FIXME: We mutate `grouped_properties` here.

        TODO: Move into a separate Facet.
        """
        choices = [
            description
            for description_property in [
                DC.description,
                SDO.description,
                RDFS.comment,
            ]
            for description in self.grouped_properties.pop(description_property, [])
            if isinstance(description, Literal)
        ]

        try:
            literal = choices[0]
        except IndexError:
            return None

        literal_value = literal.value

        match literal_value:
            case str() as string:
                return string

            case minidom.Document() as xml_document:
                return Syntax(
                    xml_document.toxml(),
                    'xml',
                )

            case something_else:
                type_of_something_else = type(something_else)
                raise ValueError(
                    f'What is this? {something_else} '   # noqa: WPS326
                    f'is a {type_of_something_else}!',   # noqa: WPS326
                )

    @functools.cached_property
    def properties(self) -> Widget | None:
        """Render properties table."""
        if not self.grouped_properties:
            return Static('No properties found ☹')

        return PropertiesContainer(*self.rows)

    def show(self) -> Widget:
        """Render the content."""

        widgets = [PageTitle(self.this)]

        if self.description:
            widgets.append(Description(self.description))
        
        widgets.append(self.properties)

        return VerticalScroll(*widgets)


class PageFooter(PageTitle):
    """Page footer."""

    DEFAULT_CSS = """
    PageFooter {
        dock: bottom;
        background: darkmagenta;
    }
    """

class InverseProperties(TextualDefaultFacet):
    """Inverse properties view."""

    META = Path(__file__).parent / 'textual-inverse-properties.yamlld'
    query_file_name = 'inverse-properties.sparql'
    properties_on_the_right = True

    def show(self) -> Widget:
        """Render the content."""
        return Vertical(
            VerticalScroll(
                self.properties,
            ),
            PageFooter(self.this, extra='[i]& its inverse RDF properties[/i]'),
        )



import inspect
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Any, Generic, Optional, TypeVar, Union

from rdflib.term import Literal, Node

from iolanta.models import NotLiteralNode
from iolanta.query_result import QueryResult, SPARQLQueryArgument

FacetOutput = TypeVar("FacetOutput")


@dataclass
class Facet(Generic[FacetOutput]):  # noqa: WPS214
    """Base facet class."""

    this: Node
    iolanta: "iolanta.Iolanta" = field(repr=False)
    as_datatype: Optional[NotLiteralNode] = None

    def __post_init__(self):
        if not isinstance(self.this, Node):
            facet_name = self.__class__.__name__
            this_type = type(self.this).__name__
            raise ValueError(
                f"Facet {facet_name} received a non-Node as this: {self.this} (type: {this_type})"
            )

    @property
    def stored_queries_path(self) -> Path:
        """Construct directory for stored queries for this facet."""
        return Path(inspect.getfile(self.__class__)).parent / "sparql"

    inference_path: Optional[Path] = None

    def query(
        self,
        query_text: str,
        **kwargs: SPARQLQueryArgument,
    ) -> QueryResult:
        """SPARQL query."""
        return self.iolanta.query(
            query_text=query_text,
            **kwargs,
        )

    def render(
        self,
        node: Union[str, Node],
        as_datatype: NotLiteralNode,
    ) -> Any:
        """Shortcut to render something via iolanta."""
        return self.iolanta.render(
            node=node,
            as_datatype=as_datatype,
        )

    def stored_query(self, file_name: str, **kwargs: SPARQLQueryArgument):
        """Execute a stored SPARQL query."""
        query_text = (self.stored_queries_path / file_name).read_text()
        return self.query(
            query_text=query_text,
            **kwargs,
        )

    def show(self) -> FacetOutput:
        """Render the facet."""
        raise NotImplementedError()

    @property
    def language(self) -> Literal:
        """Preferred language for Iolanta output."""
        return self.iolanta.language

    @cached_property
    def logger(self):
        """Logger."""
        return self.iolanta.logger.bind(facet=self.__class__.__name__)

    def __str__(self):
        """Render."""
        return str(self.show())



from enum import StrEnum
from typing import Annotated

import rich
from rdflib import Node
from rdflib.query import Result
from rich.table import Table
from typer import Option, Typer

from iolanta.sparqlspace.sparqlspace import SPARQLSpace

app = Typer()


class OutputFormat(StrEnum):
    """Output formats for the query command."""

    CSV = 'csv'      # noqa: WPS115
    JSON = 'json'    # noqa: WPS115
    TABLE = 'table'  # noqa: WPS115


def _format_node(node: Node):
    return node


def _format_result(query_result: Result, output_format: OutputFormat):
    match output_format:
        case OutputFormat.CSV:
            return query_result.serialize(format='csv').decode()

        case OutputFormat.JSON:
            return query_result.serialize(format='json').decode()

        case OutputFormat.TABLE:
            table = Table(*query_result.vars)
            for row in query_result:
                table.add_row(
                    *[
                        _format_node(node)
                        for node in row
                    ],
                )
            return table

    raise NotImplementedError(f'Output format {output_format} not implemented.')


@app.command(name='query')
def query_command(
    query: str,
    output_format: Annotated[
        OutputFormat,
        Option(help='Output format.'),
    ] = OutputFormat.TABLE,
):
    """Execute a SPARQL query."""
    query_result = SPARQLSpace().query(query)
    rich.print(
        _format_result(
            query_result=query_result,
            output_format=output_format,
        ),
    )
